from django.conf import settings

# Register all models when set to True
settings.AUDITLOG_INCLUDE_ALL_MODELS = getattr(
    settings, "AUDITLOG_INCLUDE_ALL_MODELS", False
)

# Exclude models in registration process
# It will be considered when `AUDITLOG_INCLUDE_ALL_MODELS` is True
settings.AUDITLOG_EXCLUDE_TRACKING_MODELS = getattr(
    settings, "AUDITLOG_EXCLUDE_TRACKING_MODELS", ()
)

# Register models and define their logging behaviour
settings.AUDITLOG_INCLUDE_TRACKING_MODELS = getattr(
    settings, "AUDITLOG_INCLUDE_TRACKING_MODELS", ()
)

# Exclude named fields across all models
settings.AUDITLOG_EXCLUDE_TRACKING_FIELDS = getattr(
    settings, "AUDITLOG_EXCLUDE_TRACKING_FIELDS", ()
)

# Mask named fields across all models
settings.AUDITLOG_MASK_TRACKING_FIELDS = getattr(
    settings, "AUDITLOG_MASK_TRACKING_FIELDS", ()
)

# Disable on raw save to avoid logging imports and similar
settings.AUDITLOG_DISABLE_ON_RAW_SAVE = getattr(
    settings, "AUDITLOG_DISABLE_ON_RAW_SAVE", False
)

# CID

settings.AUDITLOG_CID_HEADER = getattr(
    settings, "AUDITLOG_CID_HEADER", "x-correlation-id"
)
settings.AUDITLOG_CID_GETTER = getattr(settings, "AUDITLOG_CID_GETTER", None)

# migration
settings.AUDITLOG_TWO_STEP_MIGRATION = getattr(
    settings, "AUDITLOG_TWO_STEP_MIGRATION", False
)
settings.AUDITLOG_USE_TEXT_CHANGES_IF_JSON_IS_NOT_PRESENT = getattr(
    settings, "AUDITLOG_USE_TEXT_CHANGES_IF_JSON_IS_NOT_PRESENT", False
)

# Disable remote_addr field in database
settings.AUDITLOG_DISABLE_REMOTE_ADDR = getattr(
    settings, "AUDITLOG_DISABLE_REMOTE_ADDR", False
)

# Number of characters at which changes_display_dict property should be shown
settings.AUDITLOG_CHANGE_DISPLAY_TRUNCATE_LENGTH = getattr(
    settings, "AUDITLOG_CHANGE_DISPLAY_TRUNCATE_LENGTH", 140
)

# Use pure JSON for changes field
settings.AUDITLOG_STORE_JSON_CHANGES = getattr(
    settings, "AUDITLOG_STORE_JSON_CHANGES", False
)

settings.AUDITLOG_MASK_CALLABLE = getattr(settings, "AUDITLOG_MASK_CALLABLE", None)

settings.AUDITLOG_LOGENTRY_MODEL = getattr(
    settings, "AUDITLOG_LOGENTRY_MODEL", "auditlog.LogEntry"
)

# Use base model managers instead of default model managers
settings.AUDITLOG_USE_BASE_MANAGER = getattr(
    settings, "AUDITLOG_USE_BASE_MANAGER", False
)

# Use string representation of referenced object in foreign key changes instead of its primary key
settings.AUDITLOG_USE_FK_STRING_REPRESENTATION = getattr(
    settings, "AUDITLOG_USE_FK_STRING_REPRESENTATION", False
)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("auditlog", "0015_alter_logentry_changes"),
    ]

    operations = [
        migrations.AddField(
            model_name="logentry",
            name="remote_port",
            field=models.PositiveIntegerField(
                blank=True, null=True, verbose_name="remote port"
            ),
        ),
    ]



from urllib.parse import unquote

from django import urls as urlresolvers
from django.conf import settings
from django.contrib import admin
from django.contrib.admin.views.main import PAGE_VAR
from django.core.exceptions import PermissionDenied
from django.http import HttpRequest
from django.template.response import TemplateResponse
from django.urls import path, reverse
from django.urls.exceptions import NoReverseMatch
from django.utils.html import format_html
from django.utils.text import capfirst
from django.utils.timezone import is_aware, localtime
from django.utils.translation import gettext_lazy as _

from auditlog import get_logentry_model
from auditlog.render import get_field_verbose_name, render_logentry_changes_html
from auditlog.signals import accessed

LogEntry = get_logentry_model()

MAX = 75


class LogEntryAdminMixin:
    request: HttpRequest
    CID_TITLE = _("Click to filter by records with this correlation id")

    @admin.display(description=_("Created"))
    def created(self, obj):
        if is_aware(obj.timestamp):
            return localtime(obj.timestamp)
        return obj.timestamp

    @admin.display(description=_("User"))
    def user_url(self, obj):
        if obj.actor:
            app_label, model = settings.AUTH_USER_MODEL.split(".")
            viewname = f"admin:{app_label}_{model.lower()}_change"
            try:
                link = urlresolvers.reverse(viewname, args=[obj.actor.pk])
            except NoReverseMatch:
                return "%s" % (obj.actor)
            return format_html('<a href="{}">{}</a>', link, obj.actor)

        return "system"

    @admin.display(description=_("Resource"))
    def resource_url(self, obj):
        app_label, model = obj.content_type.app_label, obj.content_type.model
        viewname = f"admin:{app_label}_{model}_change"
        try:
            args = [obj.object_pk] if obj.object_id is None else [obj.object_id]
            link = urlresolvers.reverse(viewname, args=args)
        except NoReverseMatch:
            return obj.object_repr
        else:
            return format_html(
                '<a href="{}">{} - {}</a>', link, obj.content_type, obj.object_repr
            )

    @admin.display(description=_("Changes"))
    def msg_short(self, obj):
        if obj.action in [LogEntry.Action.DELETE, LogEntry.Action.ACCESS]:
            return ""  # delete
        changes = obj.changes_dict
        s = "" if len(changes) == 1 else "s"
        fields = ", ".join(changes.keys())
        if len(fields) > MAX:
            i = fields.rfind(" ", 0, MAX)
            fields = fields[:i] + " .."
        return "%d change%s: %s" % (len(changes), s, fields)

    @admin.display(description=_("Changes"))
    def msg(self, obj):
        return render_logentry_changes_html(obj)

    @admin.display(description="Correlation ID")
    def cid_url(self, obj):
        cid = obj.cid
        if cid:
            url = self._add_query_parameter("cid", cid)
            return format_html(
                '<a href="{}" title="{}">{}</a>', url, self.CID_TITLE, cid
            )

    def _add_query_parameter(self, key: str, value: str):
        full_path = self.request.get_full_path()
        delimiter = "&" if "?" in full_path else "?"

        return f"{full_path}{delimiter}{key}={value}"

    def field_verbose_name(self, obj, field_name: str):
        """
        Use `auditlog.render.get_field_verbose_name` instead.
        This method is kept for backward compatibility.
        """
        return get_field_verbose_name(obj, field_name)


class LogAccessMixin:
    def render_to_response(self, context, **response_kwargs):
        obj = self.get_object()
        accessed.send(obj.__class__, instance=obj)
        return super().render_to_response(context, **response_kwargs)


class AuditlogHistoryAdminMixin:
    """
    Add an audit log history view to a model admin.
    """

    auditlog_history_template = "auditlog/object_history.html"
    show_auditlog_history_link = False
    auditlog_history_per_page = 10

    def get_list_display(self, request):
        list_display = list(super().get_list_display(request))
        if self.show_auditlog_history_link and "auditlog_link" not in list_display:
            list_display.append("auditlog_link")

        return list_display

    def get_urls(self):
        opts = self.model._meta
        info = opts.app_label, opts.model_name
        my_urls = [
            path(
                "<path:object_id>/auditlog/",
                self.admin_site.admin_view(self.auditlog_history_view),
                name="%s_%s_auditlog" % info,
            )
        ]

        return my_urls + super().get_urls()

    def auditlog_history_view(self, request, object_id, extra_context=None):
        obj = self.get_object(request, unquote(object_id))
        if not self.has_view_permission(request, obj):
            raise PermissionDenied

        log_entries = (
            LogEntry.objects.get_for_object(obj)
            .select_related("actor")
            .order_by("-timestamp")
        )

        paginator = self.get_paginator(
            request, log_entries, self.auditlog_history_per_page
        )
        page_number = request.GET.get(PAGE_VAR, 1)
        page_obj = paginator.get_page(page_number)
        page_range = paginator.get_elided_page_range(page_obj.number)

        context = {
            **self.admin_site.each_context(request),
            "title": _("Audit log: %s") % obj,
            "module_name": str(capfirst(self.model._meta.verbose_name_plural)),
            "page_range": page_range,
            "page_var": PAGE_VAR,
            "pagination_required": paginator.count > self.auditlog_history_per_page,
            "object": obj,
            "opts": self.model._meta,
            "log_entries": page_obj,
            **(extra_context or {}),
        }

        return TemplateResponse(request, self.auditlog_history_template, context)

    @admin.display(description=_("Audit log"))
    def auditlog_link(self, obj):
        opts = self.model._meta
        url = reverse(
            f"admin:{opts.app_label}_{opts.model_name}_auditlog",
            args=[obj.pk],
        )

        return format_html('<a href="{}">{}</a>', url, _("View"))



from functools import wraps

from django.conf import settings

from auditlog import get_logentry_model
from auditlog.context import auditlog_disabled
from auditlog.diff import model_instance_diff
from auditlog.models import _get_manager_from_settings
from auditlog.signals import post_log, pre_log


def check_disable(signal_handler):
    """
    Decorator that passes along disabled in kwargs if any of the following is true:
    - 'auditlog_disabled' from threadlocal is true
    - raw = True and AUDITLOG_DISABLE_ON_RAW_SAVE is True
    """

    @wraps(signal_handler)
    def wrapper(*args, **kwargs):
        try:
            auditlog_disabled_value = auditlog_disabled.get()
        except LookupError:
            auditlog_disabled_value = False
        if not auditlog_disabled_value and not (
            kwargs.get("raw") and settings.AUDITLOG_DISABLE_ON_RAW_SAVE
        ):
            signal_handler(*args, **kwargs)

    return wrapper


@check_disable
def log_create(sender, instance, created, **kwargs):
    """
    Signal receiver that creates a log entry when a model instance is first saved to the database.

    Direct use is discouraged, connect your model through :py:func:`auditlog.registry.register` instead.
    """
    if created:
        _create_log_entry(
            action=get_logentry_model().Action.CREATE,
            instance=instance,
            sender=sender,
            diff_old=None,
            diff_new=instance,
            use_json_for_changes=settings.AUDITLOG_STORE_JSON_CHANGES,
        )


@check_disable
def log_update(sender, instance, **kwargs):
    """
    Signal receiver that creates a log entry when a model instance is changed and saved to the database.

    Direct use is discouraged, connect your model through :py:func:`auditlog.registry.register` instead.
    """
    if not instance._state.adding and instance.pk is not None:
        update_fields = kwargs.get("update_fields", None)
        old = _get_manager_from_settings(sender).filter(pk=instance.pk).first()
        _create_log_entry(
            action=get_logentry_model().Action.UPDATE,
            instance=instance,
            sender=sender,
            diff_old=old,
            diff_new=instance,
            fields_to_check=update_fields,
            use_json_for_changes=settings.AUDITLOG_STORE_JSON_CHANGES,
        )


@check_disable
def log_delete(sender, instance, **kwargs):
    """
    Signal receiver that creates a log entry when a model instance is deleted from the database.

    Direct use is discouraged, connect your model through :py:func:`auditlog.registry.register` instead.
    """
    if instance.pk is not None:
        _create_log_entry(
            action=get_logentry_model().Action.DELETE,
            instance=instance,
            sender=sender,
            diff_old=instance,
            diff_new=None,
            use_json_for_changes=settings.AUDITLOG_STORE_JSON_CHANGES,
        )


def log_access(sender, instance, **kwargs):
    """
    Signal receiver that creates a log entry when a model instance is accessed in a AccessLogDetailView.

    Direct use is discouraged, connect your model through :py:func:`auditlog.registry.register` instead.
    """
    if instance.pk is not None:
        _create_log_entry(
            action=get_logentry_model().Action.ACCESS,
            instance=instance,
            sender=sender,
            diff_old=None,
            diff_new=None,
            force_log=True,
            use_json_for_changes=settings.AUDITLOG_STORE_JSON_CHANGES,
        )


def _create_log_entry(
    action,
    instance,
    sender,
    diff_old,
    diff_new,
    fields_to_check=None,
    force_log=False,
    use_json_for_changes=False,
):
    pre_log_results = pre_log.send(
        sender,
        instance=instance,
        action=action,
    )

    if any(item[1] is False for item in pre_log_results):
        return
    LogEntry = get_logentry_model()

    error = None
    log_entry = None
    changes = None
    try:
        changes = model_instance_diff(
            diff_old,
            diff_new,
            fields_to_check=fields_to_check,
            use_json_for_changes=use_json_for_changes,
        )

        if force_log or changes:
            log_entry = LogEntry.objects.log_create(
                instance,
                action=action,
                changes=changes,
                force_log=force_log,
            )
    except BaseException as e:
        error = e
    finally:
        if log_entry or error:
            post_log.send(
                sender,
                instance=instance,
                instance_old=diff_old,
                action=action,
                error=error,
                pre_log_results=pre_log_results,
                changes=changes,
                log_entry=log_entry,
                log_created=log_entry is not None,
                use_json_for_changes=settings.AUDITLOG_STORE_JSON_CHANGES,
            )
        if error:
            raise error


def make_log_m2m_changes(field_name):
    """Return a handler for m2m_changed with field_name enclosed."""

    @check_disable
    def log_m2m_changes(signal, action, **kwargs):
        """Handle m2m_changed and call LogEntry.objects.log_m2m_changes as needed."""
        if action not in ["post_add", "post_clear", "post_remove"]:
            return
        LogEntry = get_logentry_model()

        model_manager = _get_manager_from_settings(kwargs["model"])

        if action == "post_clear":
            changed_queryset = model_manager.all()
        else:
            changed_queryset = model_manager.filter(pk__in=kwargs["pk_set"])

        if action in ["post_add"]:
            LogEntry.objects.log_m2m_changes(
                changed_queryset,
                kwargs["instance"],
                "add",
                field_name,
            )
        elif action in ["post_remove", "post_clear"]:
            LogEntry.objects.log_m2m_changes(
                changed_queryset,
                kwargs["instance"],
                "delete",
                field_name,
            )

    return log_m2m_changes



from math import ceil

from django.conf import settings
from django.core.management import CommandError, CommandParser
from django.core.management.base import BaseCommand

from auditlog import get_logentry_model

LogEntry = get_logentry_model()


class Command(BaseCommand):
    help = "Migrates changes from changes_text to json changes."
    requires_migrations_checks = True

    def add_arguments(self, parser: CommandParser):
        group = parser.add_argument_group()
        group.add_argument(
            "--check",
            action="store_true",
            help="Just check the status of the migration",
            dest="check",
        )
        group.add_argument(
            "-d",
            "--database",
            default=None,
            metavar="The database engine",
            help="If provided, the script will use native db operations. "
            "Otherwise, it will use LogEntry.objects.bulk_update",
            dest="db",
            type=str,
            choices=["postgres", "mysql", "oracle"],
        )
        group.add_argument(
            "-b",
            "--batch-size",
            default=500,
            help="Split the migration into multiple batches. If 0, then no batching will be done. "
            "When passing a -d/database, the batch value will be ignored.",
            dest="batch_size",
            type=int,
        )

    def handle(self, *args, **options):
        database = options["db"]
        batch_size = options["batch_size"]
        check = options["check"]

        if (not self.check_logs()) or check:
            return

        if database:
            result = self.migrate_using_sql(database)
            self.stdout.write(
                self.style.SUCCESS(
                    f"Updated {result} records using native database operations."
                )
            )
        else:
            result = self.migrate_using_django(batch_size)
            self.stdout.write(
                self.style.SUCCESS(f"Updated {result} records using django operations.")
            )

        self.check_logs()

    def check_logs(self):
        count = self.get_logs().count()
        if count:
            self.stdout.write(f"There are {count} records that needs migration.")
            return True

        self.stdout.write(self.style.SUCCESS("All records have been migrated."))
        if settings.AUDITLOG_USE_TEXT_CHANGES_IF_JSON_IS_NOT_PRESENT:
            var_msg = self.style.WARNING(
                "AUDITLOG_USE_TEXT_CHANGES_IF_JSON_IS_NOT_PRESENT"
            )
            self.stdout.write(f"You can now set {var_msg} to False.")

        return False

    def get_logs(self):
        return LogEntry.objects.filter(
            changes_text__isnull=False, changes__isnull=True
        ).exclude(changes_text__exact="")

    def migrate_using_django(self, batch_size):
        def _apply_django_migration(_logs) -> int:
            import json

            updated = []
            errors = []
            for log in _logs:
                try:
                    log.changes = json.loads(log.changes_text)
                except ValueError:
                    errors.append(log.id)
                else:
                    updated.append(log)

            LogEntry.objects.bulk_update(updated, fields=["changes"])
            if errors:
                self.stderr.write(
                    self.style.ERROR(
                        f"ValueError was raised while converting the logs with these ids into json."
                        f"They where not be included in this migration batch."
                        f"\n"
                        f"{errors}"
                    )
                )
            return len(updated)

        logs = self.get_logs()

        if not batch_size:
            return _apply_django_migration(logs)

        total_updated = 0
        for _ in range(ceil(logs.count() / batch_size)):
            total_updated += _apply_django_migration(self.get_logs()[:batch_size])
        return total_updated

    def migrate_using_sql(self, database):
        from django.db import connection

        def postgres():
            with connection.cursor() as cursor:
                cursor.execute(
                    f"""
                    UPDATE {LogEntry._meta.db_table}
                    SET changes="changes_text"::jsonb
                    WHERE changes_text IS NOT NULL
                        AND changes_text <> ''
                        AND changes IS NULL
                    """
                )
                return cursor.cursor.rowcount

        if database == "postgres":
            return postgres()

        raise CommandError(
            f"Migrating the records using {database} is not implemented. "
            f"Run this management command without passing a -d/--database argument."
        )



"""
PostgreSQL-specific tests for django-auditlog.
"""

from unittest import skipIf

from django.conf import settings
from django.test import TestCase
from test_app.models import PostgresArrayFieldModel


@skipIf(settings.TEST_DB_BACKEND != "postgresql", "PostgreSQL-specific test")
class PostgresArrayFieldModelTest(TestCase):
    databases = "__all__"

    def setUp(self):
        self.obj = PostgresArrayFieldModel.objects.create(
            arrayfield=[PostgresArrayFieldModel.RED, PostgresArrayFieldModel.GREEN],
        )

    @property
    def latest_array_change(self):
        return self.obj.history.latest().changes_display_dict["arrayfield"][1]

    def test_changes_display_dict_arrayfield(self):
        self.assertEqual(
            self.latest_array_change,
            "Red, Green",
            msg="The human readable text for the two choices, 'Red, Green' is displayed.",
        )
        self.obj.arrayfield = [PostgresArrayFieldModel.GREEN]
        self.obj.save()
        self.assertEqual(
            self.latest_array_change,
            "Green",
            msg="The human readable text 'Green' is displayed.",
        )
        self.obj.arrayfield = []
        self.obj.save()
        self.assertEqual(
            self.latest_array_change,
            "",
            msg="The human readable text '' is displayed.",
        )
        self.obj.arrayfield = [PostgresArrayFieldModel.GREEN]
        self.obj.save()
        self.assertEqual(
            self.latest_array_change,
            "Green",
            msg="The human readable text 'Green' is displayed.",
        )



from auditlog.middleware import AuditlogMiddleware


class CustomAuditlogMiddleware(AuditlogMiddleware):
    """
    Custom Middleware to couple the request's user role to log items.
    """

    def get_extra_data(self, request):
        context_data = super().get_extra_data(request)
        context_data["role"] = "Role 1"
        return context_data



import datetime

from django.core.management.base import BaseCommand
from django.db import connection

from auditlog import get_logentry_model

LogEntry = get_logentry_model()


class Command(BaseCommand):
    help = "Deletes all log entries from the database."

    def add_arguments(self, parser):
        parser.add_argument(
            "-y",
            "--yes",
            action="store_true",
            default=None,
            help="Continue without asking confirmation.",
            dest="yes",
        )
        parser.add_argument(
            "-b",
            "--before-date",
            default=None,
            help="Flush all entries with a timestamp before a given date (ISO 8601).",
            dest="before_date",
            type=datetime.date.fromisoformat,
        )
        parser.add_argument(
            "-t",
            "--truncate",
            action="store_true",
            default=None,
            help="Truncate log entry table.",
            dest="truncate",
        )

    def handle(self, *args, **options):
        answer = options["yes"]
        truncate = options["truncate"]
        before = options["before_date"]
        if truncate and before:
            self.stdout.write(
                "Truncate deletes all log entries and can not be passed with before-date."
            )
            return
        if answer is None:
            warning_message = (
                "This action will clear all log entries from the database."
            )
            if before is not None:
                warning_message = f"This action will clear all log entries before {before} from the database."
            self.stdout.write(warning_message)
            response = (
                input("Are you sure you want to continue? [y/N]: ").lower().strip()
            )
            answer = response == "y"

        if not answer:
            self.stdout.write("Aborted.")
            return

        if not truncate:
            entries = LogEntry.objects.all()
            if before is not None:
                entries = entries.filter(timestamp__date__lt=before)
            count, _ = entries.delete()
            self.stdout.write("Deleted %d objects." % count)
        else:
            database_vendor = connection.vendor
            database_display_name = connection.display_name
            table_name = LogEntry._meta.db_table
            if not TruncateQuery.support_truncate_statement(database_vendor):
                self.stdout.write(
                    "Database %s does not support truncate statement."
                    % database_display_name
                )
                return
            with connection.cursor() as cursor:
                query = TruncateQuery.to_sql(table_name)
                cursor.execute(query)
            self.stdout.write("Truncated log entry table.")


class TruncateQuery:
    SUPPORTED_VENDORS = ("postgresql", "mysql", "oracle", "microsoft")

    @classmethod
    def support_truncate_statement(cls, database_vendor) -> bool:
        return database_vendor in cls.SUPPORTED_VENDORS

    @staticmethod
    def to_sql(table_name) -> str:
        return f"TRUNCATE TABLE {table_name};"



from django.views.generic import DetailView

from auditlog.mixins import LogAccessMixin

from .models import SimpleModel


class SimpleModelDetailView(LogAccessMixin, DetailView):
    model = SimpleModel
    template_name = "simplemodel_detail.html"



# Generated by Django 4.2.25 on 2025-10-14 04:17

import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="CustomLogEntryModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "object_pk",
                    models.CharField(
                        db_index=True, max_length=255, verbose_name="object pk"
                    ),
                ),
                (
                    "object_id",
                    models.BigIntegerField(
                        blank=True, db_index=True, null=True, verbose_name="object id"
                    ),
                ),
                ("object_repr", models.TextField(verbose_name="object representation")),
                ("serialized_data", models.JSONField(null=True)),
                (
                    "action",
                    models.PositiveSmallIntegerField(
                        choices=[
                            (0, "create"),
                            (1, "update"),
                            (2, "delete"),
                            (3, "access"),
                        ],
                        db_index=True,
                        verbose_name="action",
                    ),
                ),
                (
                    "changes_text",
                    models.TextField(blank=True, verbose_name="change message"),
                ),
                ("changes", models.JSONField(null=True, verbose_name="change message")),
                (
                    "cid",
                    models.CharField(
                        blank=True,
                        db_index=True,
                        max_length=255,
                        null=True,
                        verbose_name="Correlation ID",
                    ),
                ),
                (
                    "remote_addr",
                    models.GenericIPAddressField(
                        blank=True, null=True, verbose_name="remote address"
                    ),
                ),
                (
                    "remote_port",
                    models.PositiveIntegerField(
                        blank=True, null=True, verbose_name="remote port"
                    ),
                ),
                (
                    "timestamp",
                    models.DateTimeField(
                        db_index=True,
                        default=django.utils.timezone.now,
                        verbose_name="timestamp",
                    ),
                ),
                (
                    "additional_data",
                    models.JSONField(
                        blank=True, null=True, verbose_name="additional data"
                    ),
                ),
                (
                    "actor_email",
                    models.CharField(
                        blank=True,
                        max_length=254,
                        null=True,
                        verbose_name="actor email",
                    ),
                ),
                ("role", models.CharField(blank=True, max_length=100, null=True)),
                (
                    "actor",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="actor",
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="contenttypes.contenttype",
                        verbose_name="content type",
                    ),
                ),
            ],
            options={
                "verbose_name": "log entry",
                "verbose_name_plural": "log entries",
                "ordering": ["-timestamp"],
                "get_latest_by": "timestamp",
                "abstract": False,
            },
        ),
    ]



import copy
from collections import defaultdict
from collections.abc import Callable, Collection, Iterable
from typing import Any

from django.apps import apps
from django.db.models import ManyToManyField, Model
from django.db.models.base import ModelBase
from django.db.models.signals import (
    ModelSignal,
    m2m_changed,
    post_delete,
    post_save,
    pre_save,
)

from auditlog.conf import settings
from auditlog.signals import accessed

DispatchUID = tuple[int, int, int]


class AuditLogRegistrationError(Exception):
    pass


class AuditlogModelRegistry:
    """
    A registry that keeps track of the models that use Auditlog to track changes.
    """

    DEFAULT_EXCLUDE_MODELS = (settings.AUDITLOG_LOGENTRY_MODEL, "admin.LogEntry")

    def __init__(
        self,
        create: bool = True,
        update: bool = True,
        delete: bool = True,
        access: bool = True,
        m2m: bool = True,
        custom: dict[ModelSignal, Callable] | None = None,
    ):
        from auditlog.receivers import log_access, log_create, log_delete, log_update

        self._registry = {}
        self._signals = {}
        self._m2m_signals = defaultdict(dict)

        if create:
            self._signals[post_save] = log_create
        if update:
            self._signals[pre_save] = log_update
        if delete:
            self._signals[post_delete] = log_delete
        if access:
            self._signals[accessed] = log_access
        self._m2m = m2m

        if custom is not None:
            self._signals.update(custom)

    def register(
        self,
        model: ModelBase = None,
        include_fields: list[str] | None = None,
        exclude_fields: list[str] | None = None,
        mapping_fields: dict[str, str] | None = None,
        mask_fields: list[str] | None = None,
        mask_callable: str | None = None,
        m2m_fields: Collection[str] | None = None,
        serialize_data: bool = False,
        serialize_kwargs: dict[str, Any] | None = None,
        serialize_auditlog_fields_only: bool = False,
    ):
        """
        Register a model with auditlog. Auditlog will then track mutations on this model's instances.

        :param model: The model to register.
        :param include_fields: The fields to include. Implicitly excludes all other fields.
        :param exclude_fields: The fields to exclude. Overrides the fields to include.
        :param mapping_fields: Mapping from field names to strings in diff.
        :param mask_fields: The fields to mask for sensitive info.
        :param mask_callable: The dotted path to a callable that will be used for masking. If not provided,
                              the default mask_callable will be used.
        :param m2m_fields: The fields to handle as many to many.
        :param serialize_data: Option to include a dictionary of the objects state in the auditlog.
        :param serialize_kwargs: Optional kwargs to pass to Django serializer
        :param serialize_auditlog_fields_only: Only fields being considered in changes will be serialized.
        """

        if include_fields is None:
            include_fields = []
        if exclude_fields is None:
            exclude_fields = []
        if mapping_fields is None:
            mapping_fields = {}
        if mask_fields is None:
            mask_fields = []
        if m2m_fields is None:
            m2m_fields = set()
        if serialize_kwargs is None:
            serialize_kwargs = {}

        if (serialize_kwargs or serialize_auditlog_fields_only) and not serialize_data:
            raise AuditLogRegistrationError(
                "Serializer options were given but the 'serialize_data' option is not "
                "set. Did you forget to set serialized_data to True?"
            )

        for fld in settings.AUDITLOG_EXCLUDE_TRACKING_FIELDS:
            exclude_fields.append(fld)

        for fld in settings.AUDITLOG_MASK_TRACKING_FIELDS:
            mask_fields.append(fld)

        def registrar(cls):
            """Register models for a given class."""
            if not issubclass(cls, Model):
                raise TypeError("Supplied model is not a valid model.")

            self._registry[cls] = {
                "include_fields": include_fields,
                "exclude_fields": exclude_fields,
                "mapping_fields": mapping_fields,
                "mask_fields": mask_fields,
                "mask_callable": mask_callable,
                "m2m_fields": m2m_fields,
                "serialize_data": serialize_data,
                "serialize_kwargs": serialize_kwargs,
                "serialize_auditlog_fields_only": serialize_auditlog_fields_only,
            }
            self._connect_signals(cls)

            # We need to return the class, as the decorator is basically
            # syntactic sugar for:
            # MyClass = auditlog.register(MyClass)
            return cls

        if model is None:
            # If we're being used as a decorator, return a callable with the
            # wrapper.
            return lambda cls: registrar(cls)
        else:
            # Otherwise, just register the model.
            registrar(model)

    def contains(self, model: ModelBase) -> bool:
        """
        Check if a model is registered with auditlog.

        :param model: The model to check.
        :return: Whether the model has been registered.
        :rtype: bool
        """
        return model in self._registry

    def unregister(self, model: ModelBase) -> None:
        """
        Unregister a model with auditlog. This will not affect the database.

        :param model: The model to unregister.
        """
        try:
            del self._registry[model]
        except KeyError:
            pass
        else:
            self._disconnect_signals(model)

    def get_models(self) -> list[ModelBase]:
        return list(self._registry.keys())

    def get_model_fields(self, model: ModelBase):
        return {
            "include_fields": list(self._registry[model]["include_fields"]),
            "exclude_fields": list(self._registry[model]["exclude_fields"]),
            "mapping_fields": dict(self._registry[model]["mapping_fields"]),
            "mask_fields": list(self._registry[model]["mask_fields"]),
            "mask_callable": self._registry[model]["mask_callable"],
        }

    def get_serialize_options(self, model: ModelBase):
        return {
            "serialize_data": bool(self._registry[model]["serialize_data"]),
            "serialize_kwargs": dict(self._registry[model]["serialize_kwargs"]),
            "serialize_auditlog_fields_only": bool(
                self._registry[model]["serialize_auditlog_fields_only"]
            ),
        }

    def _connect_signals(self, model):
        """
        Connect signals for the model.
        """
        from auditlog.receivers import make_log_m2m_changes

        for signal, receiver in self._signals.items():
            signal.connect(
                receiver,
                sender=model,
                dispatch_uid=self._dispatch_uid(signal, receiver),
            )
        if self._m2m:
            for field_name in self._registry[model]["m2m_fields"]:
                receiver = make_log_m2m_changes(field_name)
                self._m2m_signals[model][field_name] = receiver
                field = getattr(model, field_name)
                m2m_model = getattr(field, "through")

                m2m_changed.connect(
                    receiver,
                    sender=m2m_model,
                    dispatch_uid=self._m2m_dispatch_uid(m2m_changed, m2m_model),
                )

    def _disconnect_signals(self, model):
        """
        Disconnect signals for the model.
        """
        for signal, receiver in self._signals.items():
            signal.disconnect(
                sender=model, dispatch_uid=self._dispatch_uid(signal, receiver)
            )
        for field_name, receiver in self._m2m_signals[model].items():
            field = getattr(model, field_name)
            m2m_model = getattr(field, "through")
            m2m_changed.disconnect(
                sender=m2m_model,
                dispatch_uid=self._m2m_dispatch_uid(m2m_changed, m2m_model),
            )
        del self._m2m_signals[model]

    def _dispatch_uid(self, signal, receiver) -> DispatchUID:
        """Generate a dispatch_uid which is unique for a combination of self, signal, and receiver."""
        return id(self), id(signal), id(receiver)

    def _m2m_dispatch_uid(self, signal, sender) -> DispatchUID:
        """Generate a dispatch_uid which is unique for a combination of self, signal, and sender."""
        return id(self), id(signal), id(sender)

    def _get_model_classes(self, app_model: str) -> list[ModelBase]:
        try:
            try:
                app_label, model_name = app_model.split(".")
                return [apps.get_model(app_label, model_name)]
            except ValueError:
                return apps.get_app_config(app_model).get_models()
        except LookupError:
            return []

    def _get_exclude_models(
        self, exclude_tracking_models: Iterable[str]
    ) -> list[ModelBase]:
        exclude_models = [
            model
            for app_model in tuple(exclude_tracking_models)
            + self.DEFAULT_EXCLUDE_MODELS
            for model in self._get_model_classes(app_model)
        ]
        return exclude_models

    def _register_models(self, models: Iterable[str | dict[str, Any]]) -> None:
        models = copy.deepcopy(models)
        for model in models:
            if isinstance(model, str):
                for model_class in self._get_model_classes(model):
                    self.unregister(model_class)
                    self.register(model_class)
            elif isinstance(model, dict):
                appmodel = self._get_model_classes(model["model"])
                if not appmodel:
                    raise AuditLogRegistrationError(
                        f"An error was encountered while registering model '{model['model']}' - "
                        "make sure the app is registered correctly."
                    )
                model["model"] = appmodel[0]
                self.unregister(model["model"])
                self.register(**model)

    def register_from_settings(self):
        """
        Register models from settings variables
        """
        if not isinstance(settings.AUDITLOG_INCLUDE_ALL_MODELS, bool):
            raise TypeError("Setting 'AUDITLOG_INCLUDE_ALL_MODELS' must be a boolean")
        if not isinstance(settings.AUDITLOG_DISABLE_ON_RAW_SAVE, bool):
            raise TypeError("Setting 'AUDITLOG_DISABLE_ON_RAW_SAVE' must be a boolean")
        if not isinstance(settings.AUDITLOG_EXCLUDE_TRACKING_MODELS, (list, tuple)):
            raise TypeError(
                "Setting 'AUDITLOG_EXCLUDE_TRACKING_MODELS' must be a list or tuple"
            )

        if (
            not settings.AUDITLOG_INCLUDE_ALL_MODELS
            and settings.AUDITLOG_EXCLUDE_TRACKING_MODELS
        ):
            raise ValueError(
                "In order to use setting 'AUDITLOG_EXCLUDE_TRACKING_MODELS', "
                "setting 'AUDITLOG_INCLUDE_ALL_MODELS' must set to 'True'"
            )

        if (
            settings.AUDITLOG_EXCLUDE_TRACKING_FIELDS
            and not settings.AUDITLOG_INCLUDE_ALL_MODELS
        ):
            raise ValueError(
                "In order to use 'AUDITLOG_EXCLUDE_TRACKING_FIELDS', "
                "setting 'AUDITLOG_INCLUDE_ALL_MODELS' must be set to 'True'"
            )

        if (
            settings.AUDITLOG_MASK_TRACKING_FIELDS
            and not settings.AUDITLOG_INCLUDE_ALL_MODELS
        ):
            raise ValueError(
                "In order to use 'AUDITLOG_MASK_TRACKING_FIELDS', "
                "setting 'AUDITLOG_INCLUDE_ALL_MODELS' must be set to 'True'"
            )

        if not isinstance(settings.AUDITLOG_INCLUDE_TRACKING_MODELS, (list, tuple)):
            raise TypeError(
                "Setting 'AUDITLOG_INCLUDE_TRACKING_MODELS' must be a list or tuple"
            )

        if not isinstance(settings.AUDITLOG_EXCLUDE_TRACKING_FIELDS, (list, tuple)):
            raise TypeError(
                "Setting 'AUDITLOG_EXCLUDE_TRACKING_FIELDS' must be a list or tuple"
            )

        if not isinstance(settings.AUDITLOG_MASK_TRACKING_FIELDS, (list, tuple)):
            raise TypeError(
                "Setting 'AUDITLOG_MASK_TRACKING_FIELDS' must be a list or tuple"
            )

        for item in settings.AUDITLOG_INCLUDE_TRACKING_MODELS:
            if not isinstance(item, (str, dict)):
                raise TypeError(
                    "Setting 'AUDITLOG_INCLUDE_TRACKING_MODELS' items must be str or dict"
                )

            if isinstance(item, dict):
                if "model" not in item:
                    raise ValueError(
                        "Setting 'AUDITLOG_INCLUDE_TRACKING_MODELS' dict items must contain 'model' key"
                    )
                if "." not in item["model"]:
                    raise ValueError(
                        "Setting 'AUDITLOG_INCLUDE_TRACKING_MODELS' model must be in the "
                        "format <app_name>.<model_name>"
                    )

        if settings.AUDITLOG_INCLUDE_ALL_MODELS:
            exclude_models = self._get_exclude_models(
                settings.AUDITLOG_EXCLUDE_TRACKING_MODELS
            )

            for model in apps.get_models(include_auto_created=True):
                if model in exclude_models:
                    continue

                meta = model._meta
                if not meta.managed:
                    continue

                m2m_fields = [
                    m.name for m in meta.get_fields() if isinstance(m, ManyToManyField)
                ]

                exclude_fields = [
                    i.related_name
                    for i in meta.related_objects
                    if i.related_name and not i.related_model._meta.managed
                ]

                self.register(
                    model=model, m2m_fields=m2m_fields, exclude_fields=exclude_fields
                )

        if not isinstance(settings.AUDITLOG_STORE_JSON_CHANGES, bool):
            raise TypeError("Setting 'AUDITLOG_STORE_JSON_CHANGES' must be a boolean")

        self._register_models(settings.AUDITLOG_INCLUDE_TRACKING_MODELS)


auditlog = AuditlogModelRegistry()



from unittest.mock import patch

from django.contrib import admin
from django.contrib.admin.sites import AdminSite
from django.contrib.auth import get_user_model
from django.test import RequestFactory, TestCase
from test_app.models import SimpleModel

from auditlog.mixins import AuditlogHistoryAdminMixin


class TestModelAdmin(AuditlogHistoryAdminMixin, admin.ModelAdmin):
    model = SimpleModel
    auditlog_history_per_page = 5


class TestAuditlogHistoryAdminMixin(TestCase):
    def setUp(self):
        self.user = get_user_model().objects.create_user(
            username="test_admin", is_staff=True, is_superuser=True, is_active=True
        )
        self.site = AdminSite()

        self.admin = TestModelAdmin(SimpleModel, self.site)

        self.obj = SimpleModel.objects.create(text="Test object")

    def test_auditlog_history_view_requires_permission(self):
        request = RequestFactory().get("/")
        request.user = get_user_model().objects.create_user(
            username="non_staff_user", password="testpass"
        )

        with self.assertRaises(Exception):
            self.admin.auditlog_history_view(request, str(self.obj.pk))

    def test_auditlog_history_view_with_permission(self):
        request = RequestFactory().get("/")
        request.user = self.user

        response = self.admin.auditlog_history_view(request, str(self.obj.pk))

        self.assertEqual(response.status_code, 200)
        self.assertIn("log_entries", response.context_data)
        self.assertIn("object", response.context_data)
        self.assertEqual(response.context_data["object"], self.obj)

    def test_auditlog_history_view_pagination(self):
        """Test that pagination works correctly."""
        for i in range(10):
            self.obj.text = f"Updated text {i}"
            self.obj.save()

        request = RequestFactory().get("/")
        request.user = self.user

        response = self.admin.auditlog_history_view(request, str(self.obj.pk))

        self.assertTrue(response.context_data["pagination_required"])
        self.assertEqual(len(response.context_data["log_entries"]), 5)

    def test_auditlog_history_view_page_parameter(self):
        # Create more log entries by updating the object
        for i in range(10):
            self.obj.text = f"Updated text {i}"
            self.obj.save()

        request = RequestFactory().get("/?p=2")
        request.user = self.user

        response = self.admin.auditlog_history_view(request, str(self.obj.pk))

        # Should be on page 2
        self.assertEqual(response.context_data["log_entries"].number, 2)

    def test_auditlog_history_view_context_data(self):
        request = RequestFactory().get("/")
        request.user = self.user

        response = self.admin.auditlog_history_view(request, str(self.obj.pk))

        context = response.context_data
        required_keys = [
            "title",
            "module_name",
            "page_range",
            "page_var",
            "pagination_required",
            "object",
            "opts",
            "log_entries",
        ]

        for key in required_keys:
            self.assertIn(key, context)

        self.assertIn(str(self.obj), context["title"])
        self.assertEqual(context["object"], self.obj)
        self.assertEqual(context["opts"], self.obj._meta)

    def test_auditlog_history_view_extra_context(self):
        request = RequestFactory().get("/")
        request.user = self.user

        extra_context = {"extra_key": "extra_value"}
        response = self.admin.auditlog_history_view(
            request, str(self.obj.pk), extra_context
        )

        self.assertIn("extra_key", response.context_data)
        self.assertEqual(response.context_data["extra_key"], "extra_value")

    def test_auditlog_history_view_template(self):
        request = RequestFactory().get("/")
        request.user = self.user

        response = self.admin.auditlog_history_view(request, str(self.obj.pk))

        self.assertEqual(response.template_name, self.admin.auditlog_history_template)

    def test_auditlog_history_view_log_entries_ordering(self):
        self.obj.text = "First update"
        self.obj.save()
        self.obj.text = "Second update"
        self.obj.save()

        request = RequestFactory().get("/")
        request.user = self.user

        response = self.admin.auditlog_history_view(request, str(self.obj.pk))

        log_entries = list(response.context_data["log_entries"])
        self.assertGreaterEqual(log_entries[0].timestamp, log_entries[1].timestamp)

    def test_get_list_display_with_auditlog_link(self):
        self.admin.show_auditlog_history_link = True
        list_display = self.admin.get_list_display(RequestFactory().get("/"))

        self.assertIn("auditlog_link", list_display)

        self.admin.show_auditlog_history_link = False
        list_display = self.admin.get_list_display(RequestFactory().get("/"))

        self.assertNotIn("auditlog_link", list_display)

    def test_get_urls_includes_auditlog_url(self):
        urls = self.admin.get_urls()

        self.assertGreater(len(urls), 0)

        url_names = [
            url.name for url in urls if hasattr(url, "name") and url.name is not None
        ]
        auditlog_urls = [name for name in url_names if "auditlog" in name]
        self.assertGreater(len(auditlog_urls), 0)

    @patch("auditlog.mixins.reverse")
    def test_auditlog_link(self, mock_reverse):
        """Test that auditlog_link method returns correct HTML link."""
        # Mock the reverse function to return a test URL
        expected_url = f"/admin/test_app/simplemodel/{self.obj.pk}/auditlog/"
        mock_reverse.return_value = expected_url

        link_html = self.admin.auditlog_link(self.obj)

        self.assertIsInstance(link_html, str)

        self.assertIn("<a href=", link_html)
        self.assertIn("View</a>", link_html)

        self.assertIn(expected_url, link_html)

        opts = self.obj._meta
        expected_url_name = f"admin:{opts.app_label}_{opts.model_name}_auditlog"
        mock_reverse.assert_called_once_with(expected_url_name, args=[self.obj.pk])



import json
from collections.abc import Callable
from datetime import timezone

from django.conf import settings
from django.core.exceptions import FieldDoesNotExist, ObjectDoesNotExist
from django.db.models import NOT_PROVIDED, DateTimeField, ForeignKey, JSONField, Model
from django.utils import timezone as django_timezone
from django.utils.encoding import smart_str
from django.utils.module_loading import import_string

from auditlog import get_logentry_model


def track_field(field):
    """
    Returns whether the given field should be tracked by Auditlog.

    Untracked fields are many-to-many relations and relations to the Auditlog LogEntry model.

    :param field: The field to check.
    :type field: Field
    :return: Whether the given field should be tracked.
    :rtype: bool
    """

    # Do not track many to many relations
    if field.many_to_many:
        return False

    # Do not track relations to LogEntry
    if (
        getattr(field, "remote_field", None) is not None
        and field.remote_field.model == get_logentry_model()
    ):
        return False

    return True


def get_fields_in_model(instance):
    """
    Returns the list of fields in the given model instance. Checks whether to use the official
    _meta API or use the raw data. This method excludes many to many fields.

    :param instance: The model instance to get the fields for
    :type instance: Model
    :return: The list of fields for the given model (instance)
    :rtype: list
    """
    assert isinstance(instance, Model)

    return [f for f in instance._meta.get_fields() if track_field(f)]


def get_field_value(obj, field, use_json_for_changes=False):
    """
    Gets the value of a given model instance field.

    :param obj: The model instance.
    :type obj: Model
    :param field: The field you want to find the value of.
    :type field: Any
    :return: The value of the field as a string.
    :rtype: str
    """

    def get_default_value():
        """
        Attempts to get the default value for a field from the model's field definition.

        :return: The default value of the field or None
        """
        try:
            model_field = obj._meta.get_field(field.name)
            default = model_field.default
        except (AttributeError, FieldDoesNotExist):
            default = NOT_PROVIDED

        if default is NOT_PROVIDED:
            default = None
        elif callable(default):
            default = default()

        return smart_str(default) if not use_json_for_changes else default

    try:
        if isinstance(field, DateTimeField):
            # DateTimeFields are timezone-aware, so we need to convert the field
            # to its naive form before we can accurately compare them for changes.
            value = getattr(obj, field.name)
            try:
                value = field.to_python(value)
            except TypeError:
                return value
            if (
                value is not None
                and settings.USE_TZ
                and not django_timezone.is_naive(value)
            ):
                value = django_timezone.make_naive(value, timezone=timezone.utc)
        elif isinstance(field, JSONField):
            value = field.to_python(getattr(obj, field.name))
            if not use_json_for_changes:
                try:
                    value = json.dumps(value, sort_keys=True, cls=field.encoder)
                except TypeError:
                    pass
        elif (
            not settings.AUDITLOG_USE_FK_STRING_REPRESENTATION
            and (field.one_to_one or field.many_to_one)
            and hasattr(field, "rel_class")
        ):
            value = smart_str(getattr(obj, field.get_attname()), strings_only=True)
        else:
            value = getattr(obj, field.name)
            if not use_json_for_changes:
                value = smart_str(value)
                if type(value).__name__ == "__proxy__":
                    value = str(value)
    except (ObjectDoesNotExist, AttributeError):
        return get_default_value()

    return value


def is_primitive(obj) -> bool:
    """
    Checks if the given object is a primitive Python type that can be safely serialized to JSON.

    :param obj: The object to check
    :return: True if the object is a primitive type, False otherwise
    :rtype: bool
    """
    primitive_types = (type(None), bool, int, float, str, list, tuple, dict, set)
    return isinstance(obj, primitive_types)


def get_mask_function(mask_callable: str | None = None) -> Callable[[str], str]:
    """
    Get the masking function to use based on the following priority:
    1. Model-specific mask_callable if provided
    2. mask_callable from settings if configured
    3. Default mask_str function

    :param mask_callable: The dotted path to a callable that will be used for masking.
    :type mask_callable: str
    :return: A callable that takes a string and returns a masked version.
    :rtype: Callable[[str], str]
    """

    if mask_callable:
        return import_string(mask_callable)

    default_mask_callable = settings.AUDITLOG_MASK_CALLABLE
    if default_mask_callable:
        return import_string(default_mask_callable)

    return mask_str


def mask_str(value: str) -> str:
    """
    Masks the first half of the input string to remove sensitive data.

    :param value: The value to mask.
    :type value: str
    :return: The masked version of the string.
    :rtype: str
    """
    mask_limit = int(len(value) / 2)
    return "*" * mask_limit + value[mask_limit:]


def model_instance_diff(
    old: Model | None,
    new: Model | None,
    fields_to_check=None,
    use_json_for_changes=False,
):
    """
    Calculates the differences between two model instances. One of the instances may be ``None``
    (i.e., a newly created model or deleted model). This will cause all fields with a value to have
    changed (from ``None``).

    :param old: The old state of the model instance.
    :type old: Model
    :param new: The new state of the model instance.
    :type new: Model
    :param fields_to_check: An iterable of the field names to restrict the diff to, while ignoring the rest of
        the model's fields. This is used to pass the `update_fields` kwarg from the model's `save` method.
    :param use_json_for_changes: whether or not to use a JSON for changes
        (see settings.AUDITLOG_STORE_JSON_CHANGES)
    :type fields_to_check: Iterable
    :return: A dictionary with the names of the changed fields as keys and a two tuple of the old and new
            field values as value.
    :rtype: dict
    """
    from auditlog.registry import auditlog

    if not (old is None or isinstance(old, Model)):
        raise TypeError("The supplied old instance is not a valid model instance.")
    if not (new is None or isinstance(new, Model)):
        raise TypeError("The supplied new instance is not a valid model instance.")

    diff = {}

    if old is not None and new is not None:
        fields = set(old._meta.fields + new._meta.fields)
        model_fields = auditlog.get_model_fields(new._meta.model)
    elif old is not None:
        fields = set(get_fields_in_model(old))
        model_fields = auditlog.get_model_fields(old._meta.model)
    elif new is not None:
        fields = set(get_fields_in_model(new))
        model_fields = auditlog.get_model_fields(new._meta.model)
    else:
        fields = set()
        model_fields = None

    if fields_to_check:
        fields = {
            field
            for field in fields
            if (
                (isinstance(field, ForeignKey) and field.attname in fields_to_check)
                or (field.name in fields_to_check)
            )
        }

    # Check if fields must be filtered
    if (
        model_fields
        and (model_fields["include_fields"] or model_fields["exclude_fields"])
        and fields
    ):
        filtered_fields = []
        if model_fields["include_fields"]:
            filtered_fields = [
                field
                for field in fields
                if field.name in model_fields["include_fields"]
            ]
        else:
            filtered_fields = fields
        if model_fields["exclude_fields"]:
            filtered_fields = [
                field
                for field in filtered_fields
                if field.name not in model_fields["exclude_fields"]
            ]
        fields = filtered_fields

    for field in fields:
        old_value = get_field_value(old, field, use_json_for_changes)
        new_value = get_field_value(new, field, use_json_for_changes)

        if old_value != new_value:
            if model_fields and field.name in model_fields["mask_fields"]:
                mask_func = get_mask_function(model_fields.get("mask_callable"))

                diff[field.name] = (
                    mask_func(smart_str(old_value)),
                    mask_func(smart_str(new_value)),
                )
            else:
                if not use_json_for_changes:
                    diff[field.name] = (smart_str(old_value), smart_str(new_value))
                else:
                    # TODO: should we handle the case where the value is a django Model specifically?
                    #       for example, could create a list of ids for ManyToMany fields

                    # this maintains the behavior of the original code
                    if not is_primitive(old_value):
                        old_value = smart_str(old_value)
                    if not is_primitive(new_value):
                        new_value = smart_str(new_value)
                    diff[field.name] = (old_value, new_value)

    if len(diff) == 0:
        diff = None

    return diff



"""
Settings file for the Auditlog test suite.
"""

import os

DEBUG = True

SECRET_KEY = "test"

TEST_DB_BACKEND = os.getenv("TEST_DB_BACKEND", "sqlite3")

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.messages",
    "django.contrib.sessions",
    "django.contrib.admin",
    "django.contrib.staticfiles",
    "django.contrib.postgres",
    "custom_logentry_app",
    "auditlog",
    "test_app",
]


MIDDLEWARE = [
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
]

if os.environ.get("AUDITLOG_LOGENTRY_MODEL", None):
    MIDDLEWARE = MIDDLEWARE + ["auditlog.middleware.AuditlogMiddleware"]
else:
    MIDDLEWARE = MIDDLEWARE + ["middleware.CustomAuditlogMiddleware"]


if TEST_DB_BACKEND == "postgresql":
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.postgresql",
            "NAME": os.getenv(
                "TEST_DB_NAME", "auditlog" + os.environ.get("TOX_PARALLEL_ENV", "")
            ),
            "USER": os.getenv("TEST_DB_USER", "postgres"),
            "PASSWORD": os.getenv("TEST_DB_PASS", ""),
            "HOST": os.getenv("TEST_DB_HOST", "127.0.0.1"),
            "PORT": os.getenv("TEST_DB_PORT", "5432"),
        }
    }
elif TEST_DB_BACKEND == "mysql":
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.mysql",
            "NAME": os.getenv(
                "TEST_DB_NAME", "auditlog" + os.environ.get("TOX_PARALLEL_ENV", "")
            ),
            "USER": os.getenv("TEST_DB_USER", "root"),
            "PASSWORD": os.getenv("TEST_DB_PASS", ""),
            "HOST": os.getenv("TEST_DB_HOST", "127.0.0.1"),
            "PORT": os.getenv("TEST_DB_PORT", "3306"),
            "OPTIONS": {
                "charset": "utf8mb4",
                "init_command": "SET sql_mode='STRICT_TRANS_TABLES'",
            },
        }
    }
elif TEST_DB_BACKEND == "sqlite3":
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.sqlite3",
            "NAME": os.getenv(
                "TEST_DB_NAME",
                (
                    ":memory:"
                    if os.getenv("TOX_PARALLEL_ENV")
                    else "test_auditlog.sqlite3"
                ),
            ),
        }
    }
else:
    raise ValueError(f"Unsupported database backend: {TEST_DB_BACKEND}")

TEMPLATES = [
    {
        "APP_DIRS": True,
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    },
]

STATIC_URL = "/static/"

ROOT_URLCONF = "test_app.urls"

USE_TZ = True

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

AUDITLOG_LOGENTRY_MODEL = os.environ.get("AUDITLOG_LOGENTRY_MODEL", "auditlog.LogEntry")



from __future__ import annotations

from importlib.metadata import version

from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

__version__ = version("django-auditlog")


def get_logentry_model():
    model_string = getattr(settings, "AUDITLOG_LOGENTRY_MODEL", "auditlog.LogEntry")
    try:
        return django_apps.get_model(model_string, require_ready=False)
    except ValueError:
        raise ImproperlyConfigured(
            "AUDITLOG_LOGENTRY_MODEL must be of the form 'app_label.model_name'"
        )
    except LookupError:
        raise ImproperlyConfigured(
            "AUDITLOG_LOGENTRY_MODEL refers to model '%s' that has not been installed"
            % model_string
        )



from django.test import TestCase, override_settings
from test_app.models import JSONModel, NullableFieldModel, RelatedModel, SimpleModel

from auditlog import get_logentry_model
from auditlog.registry import AuditlogModelRegistry

LogEntry = get_logentry_model()


class JSONForChangesTest(TestCase):

    def setUp(self):
        self.test_auditlog = AuditlogModelRegistry()

    @override_settings(AUDITLOG_STORE_JSON_CHANGES="str")
    def test_wrong_setting_type(self):
        with self.assertRaisesMessage(
            TypeError, "Setting 'AUDITLOG_STORE_JSON_CHANGES' must be a boolean"
        ):
            self.test_auditlog.register_from_settings()

    @override_settings(AUDITLOG_STORE_JSON_CHANGES=True)
    def test_use_json_for_changes_with_simplemodel(self):
        self.test_auditlog.register_from_settings()

        smm = SimpleModel()
        smm.save()
        changes_dict = smm.history.latest().changes_dict

        # compare the id, text, boolean and datetime fields
        id_field_changes = changes_dict["id"]
        self.assertIsNone(id_field_changes[0])
        self.assertIsInstance(
            id_field_changes[1], int
        )  # the id depends on state of the database

        text_field_changes = changes_dict["text"]
        self.assertEqual(text_field_changes, [None, ""])

        boolean_field_changes = changes_dict["boolean"]
        self.assertEqual(boolean_field_changes, [None, False])

        # datetime should be serialized to string
        datetime_field_changes = changes_dict["datetime"]
        self.assertIsNone(datetime_field_changes[0])
        self.assertIsInstance(datetime_field_changes[1], str)

    @override_settings(AUDITLOG_STORE_JSON_CHANGES=True)
    def test_use_json_for_changes_with_jsonmodel(self):
        self.test_auditlog.register_from_settings()

        json_model = JSONModel()
        json_model.json = {"test_key": "test_value"}
        json_model.save()
        changes_dict = json_model.history.latest().changes_dict

        id_field_changes = changes_dict["json"]
        self.assertEqual(id_field_changes, [None, {"test_key": "test_value"}])

    @override_settings(AUDITLOG_STORE_JSON_CHANGES=True)
    def test_use_json_for_changes_with_jsonmodel_with_empty_list(self):
        self.test_auditlog.register_from_settings()

        json_model = JSONModel()
        json_model.json = []
        json_model.save()
        changes_dict = json_model.history.latest().changes_dict

        id_field_changes = changes_dict["json"]
        self.assertEqual(id_field_changes, [None, []])

    @override_settings(AUDITLOG_STORE_JSON_CHANGES=True)
    def test_use_json_for_changes_with_jsonmodel_with_complex_data(self):
        self.test_auditlog.register_from_settings()

        json_model = JSONModel()
        json_model.json = {
            "key": "test_value",
            "key_dict": {"inner_key": "inner_value"},
            "key_tuple": ("item1", "item2", "item3"),
        }
        json_model.save()
        changes_dict = json_model.history.latest().changes_dict

        id_field_changes = changes_dict["json"]
        self.assertEqual(
            id_field_changes,
            [
                None,
                {
                    "key": "test_value",
                    "key_dict": {"inner_key": "inner_value"},
                    "key_tuple": [
                        "item1",
                        "item2",
                        "item3",
                    ],  # tuple is converted to list, that's ok
                },
            ],
        )

    @override_settings(AUDITLOG_STORE_JSON_CHANGES=True)
    def test_use_json_for_changes_with_jsonmodel_with_related_model(self):
        self.test_auditlog.register_from_settings()

        simple = SimpleModel.objects.create()
        one_simple = SimpleModel.objects.create()
        related_model = RelatedModel.objects.create(
            one_to_one=simple, related=one_simple
        )
        related_model.save()
        changes_dict = related_model.history.latest().changes_dict

        field_related_changes = changes_dict["related"]
        self.assertEqual(field_related_changes, [None, one_simple.id])

        field_one_to_one_changes = changes_dict["one_to_one"]
        self.assertEqual(field_one_to_one_changes, [None, simple.id])

    @override_settings(AUDITLOG_STORE_JSON_CHANGES=True)
    def test_use_json_for_changes_update(self):
        self.test_auditlog.register_from_settings()

        simple = SimpleModel(text="original")
        simple.save()
        simple.text = "new"
        simple.save()

        changes_dict = simple.history.latest().changes_dict

        text_changes = changes_dict["text"]
        self.assertEqual(text_changes, ["original", "new"])

    @override_settings(AUDITLOG_STORE_JSON_CHANGES=True)
    def test_use_json_for_changes_delete(self):
        self.test_auditlog.register_from_settings()

        simple = SimpleModel()
        simple.save()
        simple.delete()

        history = LogEntry.objects.all()

        self.assertEqual(history.count(), 1, '"DELETE" record is always retained')

        changes_dict = history.first().changes_dict

        self.assertTrue(
            all(v[1] is None for k, v in changes_dict.items()),
            'all values in the changes dict should None, not "None"',
        )

    @override_settings(AUDITLOG_STORE_JSON_CHANGES=False)
    def test_nullable_field_with_none_not_logged(self):
        self.test_auditlog.register_from_settings()

        obj = NullableFieldModel.objects.create(time=None, optional_text=None)
        changes_dict = obj.history.latest().changes_dict

        # None → None should NOT be logged as a change
        self.assertNotIn("time", changes_dict)
        self.assertNotIn("optional_text", changes_dict)

    @override_settings(AUDITLOG_STORE_JSON_CHANGES=False)
    def test_nullable_field_with_value_logged(self):
        self.test_auditlog.register_from_settings()

        obj = NullableFieldModel.objects.create(optional_text="something")
        changes_dict = obj.history.latest().changes_dict

        # None → "something" should be logged
        self.assertIn("optional_text", changes_dict)
        self.assertEqual(changes_dict["optional_text"], ["None", "something"])

    @override_settings(AUDITLOG_STORE_JSON_CHANGES=True)
    def test_nullable_field_with_none_not_logged_json_mode(self):
        self.test_auditlog.register_from_settings()

        obj = NullableFieldModel.objects.create(time=None, optional_text=None)
        changes_dict = obj.history.latest().changes_dict

        # None → None should NOT be logged
        self.assertNotIn("time", changes_dict)
        self.assertNotIn("optional_text", changes_dict)

    @override_settings(AUDITLOG_STORE_JSON_CHANGES=False)
    def test_nullable_field_update_none_to_value(self):
        self.test_auditlog.register_from_settings()

        obj = NullableFieldModel.objects.create(optional_text=None)
        obj.optional_text = "updated"
        obj.save()

        changes_dict = obj.history.latest().changes_dict

        # None → "updated" should be logged
        self.assertIn("optional_text", changes_dict)
        self.assertEqual(changes_dict["optional_text"], ["None", "updated"])



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("auditlog", "0016_logentry_remote_port"),
    ]

    operations = [
        migrations.AddField(
            model_name="logentry",
            name="actor_email",
            field=models.CharField(
                null=True,
                verbose_name="actor email",
                blank=True,
                max_length=254,
            ),
        ),
    ]



import contextlib
import time
from contextvars import ContextVar
from functools import partial

from django.contrib.auth import get_user_model
from django.db.models.signals import pre_save

from auditlog import get_logentry_model

auditlog_value = ContextVar("auditlog_value")
auditlog_disabled = ContextVar("auditlog_disabled", default=False)


@contextlib.contextmanager
def set_actor(actor, remote_addr=None, remote_port=None):
    context_data = {
        "actor": actor,
        "remote_addr": remote_addr,
        "remote_port": remote_port,
    }
    return call_context_manager(context_data)


@contextlib.contextmanager
def set_extra_data(context_data):
    return call_context_manager(context_data)


def call_context_manager(context_data):
    """Connect a signal receiver with current user attached."""
    LogEntry = get_logentry_model()
    # Initialize thread local storage
    context_data["signal_duid"] = ("set_actor", time.time())
    auditlog_value.set(context_data)

    # Connect signal for automatic logging
    set_extra_data = partial(
        _set_extra_data,
        signal_duid=context_data["signal_duid"],
    )
    pre_save.connect(
        set_extra_data,
        sender=LogEntry,
        dispatch_uid=context_data["signal_duid"],
        weak=False,
    )

    try:
        yield
    finally:
        try:
            auditlog = auditlog_value.get()
        except LookupError:
            pass
        else:
            pre_save.disconnect(sender=LogEntry, dispatch_uid=auditlog["signal_duid"])


def _set_actor(auditlog, instance, sender):
    LogEntry = get_logentry_model()
    auth_user_model = get_user_model()
    if "actor" in auditlog:
        actor = auditlog.get("actor")
        if (
            sender == LogEntry
            and isinstance(actor, auth_user_model)
            and instance.actor is None
        ):
            instance.actor = actor
            instance.actor_email = getattr(actor, "email", None)


def _set_extra_data(sender, instance, signal_duid, **kwargs):
    """Signal receiver with extra 'user' and 'signal_duid' kwargs.

    This function becomes a valid signal receiver when it is curried with the actor and a dispatch id.
    """
    LogEntry = get_logentry_model()
    try:
        auditlog = auditlog_value.get()
    except LookupError:
        pass
    else:
        if signal_duid != auditlog["signal_duid"]:
            return

        _set_actor(auditlog, instance, sender)

        for key in auditlog:
            if key != "actor" and hasattr(LogEntry, key):
                if callable(auditlog[key]):
                    setattr(instance, key, auditlog[key]())
                else:
                    setattr(instance, key, auditlog[key])


@contextlib.contextmanager
def disable_auditlog():
    token = auditlog_disabled.set(True)
    try:
        yield
    finally:
        try:
            auditlog_disabled.reset(token)
        except LookupError:
            pass



from django.test import TestCase
from test_app.models import SimpleModel

from auditlog import get_logentry_model
from auditlog.templatetags.auditlog_tags import render_logentry_changes_html

LogEntry = get_logentry_model()


class RenderChangesTest(TestCase):

    def _create_log_entry(self, action, changes):
        return LogEntry.objects.log_create(
            SimpleModel.objects.create(),
            action=action,
            changes=changes,
        )

    def test_render_changes_empty(self):
        log_entry = self._create_log_entry(LogEntry.Action.CREATE, {})
        result = render_logentry_changes_html(log_entry)
        self.assertEqual(result, "")

    def test_render_changes_simple_field(self):
        changes = {"text": ["old text", "new text"]}
        log_entry = self._create_log_entry(LogEntry.Action.UPDATE, changes)

        result = render_logentry_changes_html(log_entry)

        self.assertIn("<table>", result)
        self.assertIn("<th>#</th>", result)
        self.assertIn("<th>Field</th>", result)
        self.assertIn("<th>From</th>", result)
        self.assertIn("<th>To</th>", result)
        self.assertIn("old text", result)
        self.assertIn("new text", result)
        self.assertIsInstance(result, str)

    def test_render_changes_password_field(self):
        changes = {"password": ["oldpass", "newpass"]}
        log_entry = self._create_log_entry(LogEntry.Action.UPDATE, changes)

        result = render_logentry_changes_html(log_entry)

        self.assertIn("***", result)
        self.assertNotIn("oldpass", result)
        self.assertNotIn("newpass", result)

    def test_render_changes_m2m_field(self):
        changes = {
            "related_objects": {
                "type": "m2m",
                "operation": "add",
                "objects": ["obj1", "obj2", "obj3"],
            }
        }
        log_entry = self._create_log_entry(LogEntry.Action.UPDATE, changes)

        result = render_logentry_changes_html(log_entry)

        self.assertIn("<table>", result)
        self.assertIn("<th>#</th>", result)
        self.assertIn("<th>Relationship</th>", result)
        self.assertIn("<th>Action</th>", result)
        self.assertIn("<th>Objects</th>", result)
        self.assertIn("add", result)
        self.assertIn("obj1", result)
        self.assertIn("obj2", result)
        self.assertIn("obj3", result)

    def test_render_changes_mixed_fields(self):
        changes = {
            "text": ["old text", "new text"],
            "related_objects": {
                "type": "m2m",
                "operation": "remove",
                "objects": ["obj1"],
            },
        }
        log_entry = self._create_log_entry(LogEntry.Action.UPDATE, changes)

        result = render_logentry_changes_html(log_entry)

        tables = result.count("<table>")
        self.assertEqual(tables, 2)

        self.assertIn("old text", result)
        self.assertIn("new text", result)

        self.assertIn("remove", result)
        self.assertIn("obj1", result)

    def test_render_changes_field_verbose_name(self):
        changes = {"text": ["old", "new"]}
        log_entry = self._create_log_entry(LogEntry.Action.UPDATE, changes)

        result = render_logentry_changes_html(log_entry)

        self.assertIn("Text", result)

    def test_render_changes_with_none_values(self):
        changes = {"text": [None, "new text"], "boolean": [True, None]}
        log_entry = self._create_log_entry(LogEntry.Action.UPDATE, changes)

        result = render_logentry_changes_html(log_entry)

        self.assertIn("None", result)
        self.assertIn("new text", result)
        self.assertIn("True", result)

    def test_render_changes_sorted_fields(self):
        changes = {
            "z_field": ["old", "new"],
            "a_field": ["old", "new"],
            "m_field": ["old", "new"],
        }
        log_entry = self._create_log_entry(LogEntry.Action.UPDATE, changes)

        result = render_logentry_changes_html(log_entry)

        a_index = result.find("A field")
        m_index = result.find("M field")
        z_index = result.find("Z field")

        self.assertLess(a_index, m_index)
        self.assertLess(m_index, z_index)

    def test_render_changes_m2m_sorted_fields(self):
        changes = {
            "z_related": {"type": "m2m", "operation": "add", "objects": ["obj1"]},
            "a_related": {"type": "m2m", "operation": "remove", "objects": ["obj2"]},
        }
        log_entry = self._create_log_entry(LogEntry.Action.UPDATE, changes)

        result = render_logentry_changes_html(log_entry)

        a_index = result.find("A related")
        z_index = result.find("Z related")

        self.assertLess(a_index, z_index)

    def test_render_changes_create_action(self):
        changes = {
            "text": [None, "new value"],
            "boolean": [None, True],
        }
        log_entry = self._create_log_entry(LogEntry.Action.CREATE, changes)

        result = render_logentry_changes_html(log_entry)

        self.assertIn("<table>", result)
        self.assertIn("new value", result)
        self.assertIn("True", result)

    def test_render_changes_delete_action(self):
        changes = {
            "text": ["old value", None],
            "boolean": [True, None],
        }
        log_entry = self._create_log_entry(LogEntry.Action.DELETE, changes)

        result = render_logentry_changes_html(log_entry)

        self.assertIn("<table>", result)
        self.assertIn("old value", result)
        self.assertIn("True", result)
        self.assertIn("None", result)



from django import template

from auditlog.render import render_logentry_changes_html as render_changes

register = template.Library()


@register.filter
def render_logentry_changes_html(log_entry):
    """
    Format LogEntry changes as HTML.

    Usage in template:
    {{ log_entry_object|render_logentry_changes_html|safe }}
    """
    return render_changes(log_entry)



from django.core.exceptions import FieldDoesNotExist
from django.forms.utils import pretty_name
from django.utils.html import format_html, format_html_join
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _


def render_logentry_changes_html(log_entry):
    changes = log_entry.changes_dict
    if not changes:
        return ""

    atom_changes = {}
    m2m_changes = {}

    # Separate regular fields from M2M changes
    for field, change in changes.items():
        if isinstance(change, dict) and change.get("type") == "m2m":
            m2m_changes[field] = change
        else:
            atom_changes[field] = change

    html_parts = []

    # Render regular field changes
    if atom_changes:
        html_parts.append(_render_field_changes(log_entry, atom_changes))

    # Render M2M relationship changes
    if m2m_changes:
        html_parts.append(_render_m2m_changes(log_entry, m2m_changes))

    return mark_safe("".join(html_parts))


def get_field_verbose_name(log_entry, field_name):
    from auditlog.registry import auditlog

    model = log_entry.content_type.model_class()
    if model is None:
        return field_name

    # Try to get verbose name from auditlog mapping
    try:
        if auditlog.contains(model._meta.model):
            model_fields = auditlog.get_model_fields(model._meta.model)
            mapping_field_name = model_fields["mapping_fields"].get(field_name)
            if mapping_field_name:
                return mapping_field_name
    except KeyError:
        # Model definition in auditlog was probably removed
        pass

    # Fall back to Django field verbose_name
    try:
        field = model._meta.get_field(field_name)
        return pretty_name(getattr(field, "verbose_name", field_name))
    except FieldDoesNotExist:
        return pretty_name(field_name)


def _render_field_changes(log_entry, atom_changes):
    rows = []
    rows.append(_format_header("#", _("Field"), _("From"), _("To")))

    for i, (field, change) in enumerate(sorted(atom_changes.items()), 1):
        field_name = get_field_verbose_name(log_entry, field)
        values = ["***", "***"] if field == "password" else change
        rows.append(_format_row(i, field_name, *values))

    return f"<table>{''.join(rows)}</table>"


def _render_m2m_changes(log_entry, m2m_changes):
    rows = []
    rows.append(_format_header("#", _("Relationship"), _("Action"), _("Objects")))

    for i, (field, change) in enumerate(sorted(m2m_changes.items()), 1):
        field_name = get_field_verbose_name(log_entry, field)
        objects_html = format_html_join(
            mark_safe("<br>"),
            "{}",
            [(obj,) for obj in change["objects"]],
        )
        rows.append(_format_row(i, field_name, change["operation"], objects_html))

    return f"<table>{''.join(rows)}</table>"


def _format_header(*labels):
    return format_html("".join(["<tr>", "<th>{}</th>" * len(labels), "</tr>"]), *labels)


def _format_row(*values):
    return format_html("".join(["<tr>", "<td>{}</td>" * len(values), "</tr>"]), *values)



import django.dispatch

accessed = django.dispatch.Signal()


pre_log = django.dispatch.Signal()
"""
Whenever an audit log entry is written, this signal
is sent before writing the log.
Keyword arguments sent with this signal:

:param class sender:
    The model class that's being audited.

:param Any instance:
    The actual instance that's being audited.

:param Action action:
    The action on the model resulting in an
    audit log entry. Type: :class:`auditlog.models.LogEntry.Action`

The receivers' return values are sent to any :func:`post_log`
signal receivers, with one exception: if any receiver returns False,
no logging will be made. This can be useful if logging should be
conditionally enabled / disabled
"""

post_log = django.dispatch.Signal()
"""
Whenever an audit log entry is written, this signal
is sent after writing the log.
This signal is also fired when there is an error in creating the log.

Keyword arguments sent with this signal:

:param class sender:
    The model class that's being audited.

:param Any instance:
    The actual instance that's being audited.

:param Action action:
    The action on the model resulting in an
    audit log entry. Type: :class:`auditlog.models.LogEntry.Action`

:param Optional[dict] changes:
    The changes that were logged. If there was en error while determining the changes,
    this will be None. In some cases, such as when logging access to the instance,
    the changes will be an empty dict.

:param Optional[LogEntry] log_entry:
    The log entry that was created and stored in the database. If there was an error,
    this will be None.

:param bool log_created:
    Was the log actually created?
    This could be false if there was an error in creating the log.

:param Optional[Exception] error:
    The error, if one occurred while saving the audit log entry. ``None``,
    otherwise

:param List[Tuple[method,Any]] pre_log_results:
    List of tuple pairs ``[(pre_log_receiver, pre_log_response)]``, where
    ``pre_log_receiver`` is the receiver method, and ``pre_log_response`` is the
    corresponding response of that method. If there are no :const:`pre_log` receivers,
    then the list will be empty. ``pre_log_receiver`` is guaranteed to be
    non-null, but ``pre_log_response`` may be ``None``. This depends on the corresponding
    ``pre_log_receiver``'s return value.
"""
